import logging
from typing import Optional, Union

from aiohttp import ClientError, ClientResponseError, ClientSession
from aiohttp.hdrs import AUTHORIZATION, USER_AGENT

from .consts import (
    USER_AGENT_PRODUCT,
    USER_AGENT_PRODUCT_VERSION,
    USER_AGENT_SYSTEM_INFORMATION,
)
from .endpoints import (
    API_URL_SCAFFOLD,
    APP_MUST_UPDATE,
    AUTH_URL_SCAFFOLD,
    DATA_HOUR,
    DATA_LATEST,
    DEVICE_DATA,
    USER_CONFIG,
    USER_LOGIN,
    USER_REFRESH_TOKEN,
    USER_VERIFY_EMAIL,
)
from .errors import ForbiddenError, RequestError, UnauthorizedError
from .util import json_pp


class API(object):
    """uHoo API"""

    def __init__(self, websession: ClientSession) -> None:
        self._log: logging.Logger = logging.getLogger("pyuhoo")

        self._user_agent: str = (
            f"{USER_AGENT_PRODUCT}"
            + "/"
            + f"{USER_AGENT_PRODUCT_VERSION} "
            + f"({USER_AGENT_SYSTEM_INFORMATION})"
        )

        self._websession: ClientSession = websession

        self._bearer_token: Optional[str] = None

    async def _request(
        self, method: str, scaffold: str, endpoint: str, data: Optional[dict] = None
    ):
        headers = {}
        if self._bearer_token:
            headers.update({AUTHORIZATION: f"Bearer {self._bearer_token}"})
        if self._user_agent:
            headers.update({USER_AGENT: self._user_agent})

        self._log.debug(f"[_request] {method} {scaffold}/{endpoint}")

        if method.lower() == "post":
            self._log.debug(f"[_request] {json_pp(data)}")

        async with self._websession.request(
            method, f"{scaffold}/{endpoint}", headers=headers, data=data
        ) as resp:
            try:
                self._log.debug(
                    f"[_request] {resp.status} {method} {scaffold}/{endpoint}"
                )

                if resp.content_type == "application/json":
                    json = await resp.json()
                else:
                    text = await resp.text()

                resp.raise_for_status()

                return json

            except ClientResponseError as err:
                if err.status == 401:
                    self._log.debug(f"[_request] 401 Unauthorized:\n{json_pp(json)}")
                    raise UnauthorizedError(json_pp(json)) from None
                elif err.status == 403:
                    self._log.debug(f"[_request] 403 Unauthorized:\n{text}")
                    raise ForbiddenError(text) from None
                else:
                    raise RequestError(
                        f"Error requesting data from {scaffold}/{endpoint}: {err}"
                    ) from None

            except ClientError as err:
                raise RequestError(
                    f"Error requesting data from {scaffold}/{endpoint}: {err}"
                ) from None

    def set_bearer_token(self, bearer_token: Optional[str]) -> None:
        self._bearer_token = bearer_token

    async def app_must_update(self, version: int) -> bool:
        resp: Union[int, dict] = await self._request(
            "post", API_URL_SCAFFOLD, APP_MUST_UPDATE, data={"version": version}
        )
        if resp == 0:
            return False
        else:
            self._log.debug(f"[app_must_update] received non-zero response: {resp}")
            return True

    async def user_config(self) -> dict:
        resp: dict = await self._request("get", AUTH_URL_SCAFFOLD, USER_CONFIG)
        return resp

    async def user_verify_email(self, username: str, client_id: str) -> dict:
        resp: dict = await self._request(
            "post",
            AUTH_URL_SCAFFOLD,
            USER_VERIFY_EMAIL,
            data={"username": username, "clientId": client_id},
        )
        return resp

    async def user_login(self, username: str, password: str, client_id: str) -> dict:
        """Note: password is an encrypted hash of the user's password"""
        resp: dict = await self._request(
            "post",
            AUTH_URL_SCAFFOLD,
            USER_LOGIN,
            data={"username": username, "password": password, "clientId": client_id},
        )
        return resp

    async def user_refresh_token(self, token, user_device_id) -> dict:
        """Note: user_device_id is the same as client_id"""
        resp: dict = await self._request(
            "post",
            AUTH_URL_SCAFFOLD,
            USER_REFRESH_TOKEN,
            data={"Token": token, "userDeviceId": user_device_id},
        )
        return resp

    async def data_latest(self) -> dict:
        resp: dict = await self._request("get", API_URL_SCAFFOLD, DATA_LATEST)
        return resp

    async def data_hour(self, serial_number: str, prev_date_time: str) -> dict:
        resp: dict = await self._request(
            "post",
            API_URL_SCAFFOLD,
            DATA_HOUR,
            data={"serialNumber": serial_number, "prevDateTime": prev_date_time},
        )
        return resp

    async def device_data(self, serial_number: str) -> dict:
        resp: dict = await self._request(
            "post", API_URL_SCAFFOLD, DEVICE_DATA, data={"serialNumber": serial_number}
        )
        return resp



# URL endpoints and paths taken from uHoo Android app

# URL Endpoints
API_URL_SCAFFOLD: str = "https://api.uhooinc.com/v1"
AUTH_URL_SCAFFOLD: str = "https://auth.uhooinc.com"

# API Paths
DEVICE_PAIR: str = "pairdevice"
USER_REFRESH_TOKEN: str = "renewusertoken"
USER_CONFIG: str = "user"
USER_VERIFY_EMAIL: str = "verifyemail"
USER_LOGIN: str = "login"
APP_MUST_UPDATE: str = "getandroidversionconsumer"
APP_MUST_UPDATE_BIZ: str = "getandroidversion"
USER_LOGOUT: str = "clearusersession"
USER_REGISTER: str = "registeruser"
USER_EDIT_DATA: str = "edituserdata"
USER_DATA: str = "getuserdata"
USER_AVATAR: str = "useravatar"
USER_SET_AVATAR: str = "useravatar"
USER_FORGOT_PASSWORD: str = "forgetpw"
USER_RESET_PASSWORD: str = "resetpw"
USER_CHANGE_PASSWORD: str = "changepw"
USER_LANGUAGES: str = "languagelist"
USER_VERIFY: str = "getuserinformation"
USER_SETTINGS: str = "wtvsRh/usersettings"
USER_SET_SETTINGS: str = "usersettings"
DEVICE_ALL: str = "getdevice"
DEVICE_DATA: str = "wtvsRh/getdevicedata"
DEVICE_GET_DETAILS: str = "getdevicedetails"
DEVICE_SAVE_DETAILS: str = "savedeviceconsumer"
DEVICE_SAVE_CORPORATE: str = "test/savecorporatedevice"
DEVICE_EDIT_DETAILS: str = "test/editdevicedetails"
DEVICE_CHECK_MAC: str = "checkproddevice"
DEVICE_IS_REGISTERED: str = "device/whoami"
DEVICE_REGIONS: str = "deviceregion"
DEVICE_TIMEZONES: str = "gettimezones"
DEVICE_ROOMS: str = "getlocationtypes"
DEVICE_TRANSFER_OWNER: str = "transferowner"
DATA_LATEST: str = "getalllatestdata"
DATA_HOUR: str = "wtvsRh/gethourcolor"
DATA_DAY: str = "wtvsRh/getdaycolor"
DATA_WEEK: str = "wtvsRh/getweekcolor"
DATA_MONTH: str = "wtvsRh/getmonthcolor"
DATA_DOWNLOAD_HOUR: str = "wtvsRh/datahouravg"
DATA_DOWNLOAD_MINUTE: str = "wtvsRh/dataminute"
SHARE_TO_LIST: str = "sharetolist"
SHARE_TO_LIST_BIZ: str = "sharetolistbiz"
SHARE_DEVICE: str = "autosharedevice"
SHARE_DEVICE_SUPERADMIN: str = "sharedevicesa"
SHARE_DEVICE_ADMIN: str = "sharedeviceadmin"
SHARE_UNSHARE_DEVICE: str = "unsharedevice"
SHARE_UNSHARE_DEVICE_SUPERADMIN: str = "unsharedevicesa"
SHARE_UNSHARE_DEVICE_ADMIN: str = "unsharedeviceadmin"
SHARE_ACCEPT_DEVICE: str = "acceptsharedevice"
NOTIFS_ALL: str = "wtvsRh/allnotif"
NOTIFS_NEXT: str = "wtvsRh/allnotif"
NOTIFS_DEVICE: str = "wtvsRh/devicenotif"
NOTIFS_DEVICE_NEXT: str = "wtvsRh/devicenotif"
NOTIFS_DELETE: str = "deletedevicenotif"
NOTIFS_DEVICE_DELETE: str = "deletedevicenotif"
NOTIFS_SYSTEM: str = "systemnotifs"
NOTIFS_SETTINGS: str = "getnotifsettings"
NOTIFS_SETTINGS_SET: str = "notifsettings"
NOTIFS_SETTINGS_SET_QUEUE: str = "notifsettings"
THRESHOLDS_DEFAULT: str = "wtvsRh/thresholddefault"
THRESHOLDS_DEVICE: str = "wtvsRh/getdevicethreshold"
THRESHOLDS_DEVICE_EDIT: str = "thresholdupdateall"
SHOP_GET_ITEMS: str = "uhoolist"
PUSH_SET_DEVICE: str = "pushdevicecontrol"
PUSH_TEST: str = "pushtest"
INTEGRATE_IFTTT_USERTOKEN: str = "ifttttoken"



"""Define package errors."""


class UhooError(Exception):
    """Define a base error."""

    pass


class RequestError(UhooError):
    """Define an error related to invalid requests."""

    pass


class UnauthorizedError(UhooError):
    """Define an error for 401 unauthorized responses"""

    pass


class ForbiddenError(UhooError):
    """Define an error for 403 forbidden responses"""

    pass



"""Define various utility functions."""
import json
from codecs import decode, encode
from hashlib import md5, sha256
from typing import Any, cast

from Crypto.Cipher import AES
from Crypto.Cipher._mode_ecb import EcbMode
from Crypto.Util.Padding import pad


def json_pp(obj: Any) -> str:
    return json.dumps(obj, sort_keys=True, indent=2, separators=(",", ": "))


def salted_hash(password: str, u_id: str) -> str:
    salt: bytes = b"@uhooinc.com"
    salted_hash = sha256()
    salted_hash.update(bytes(u_id, "utf-8") + bytes(password, "utf-8") + salt)
    hexdigest: str = salted_hash.hexdigest()
    return hexdigest


def encrypted_hash(code: str, pt_hash: str) -> str:
    code_hash = md5()
    code_hash.update(bytes(code, "utf-8"))
    code_hash_hex: str = code_hash.hexdigest()

    key: bytes = decode(bytes(code_hash_hex, "utf-8"), "hex")

    pt_hash_padded: bytes = pad(bytes(pt_hash, "utf-8"), 16)

    cipher: EcbMode = cast(EcbMode, AES.new(key, AES.MODE_ECB))
    ct: bytes = cipher.encrypt(pt_hash_padded)
    return str(encode(ct, "hex"), "utf-8")



import logging
from typing import Dict, Optional

from aiohttp import ClientSession

from pyuhoo.errors import ForbiddenError, UhooError, UnauthorizedError

from .api import API
from .consts import APP_VERSION, CLIENT_ID
from .device import Device
from .util import encrypted_hash, json_pp, salted_hash


class Client(object):
    def __init__(
        self, username: str, password: str, websession: ClientSession, **kwargs
    ) -> None:
        self._log: logging.Logger = logging.getLogger("pyuhoo")

        if kwargs.get("debug") is True:
            self._log.setLevel(logging.DEBUG)
            self._log.debug("Debug mode is explicitly enabled.")
        else:
            self._log.debug(
                "Debug mode is not explicitly enabled (but may be enabled elsewhere)."
            )

        self._app_version: int = APP_VERSION
        self._client_id: str = CLIENT_ID
        self._device_id: Optional[str] = None
        self._devices: Dict[str, Device] = {}
        self._username: str = username
        self._password: str = password
        self._websession: ClientSession = websession
        self._token: Optional[str] = None
        self._refresh_token: Optional[str] = None
        self.user_settings_temp: str = "f"  # "f" or "c"

        self._api: API = API(self._websession)

    async def login(self) -> None:
        # app_must_update: bool = await self._api.app_must_update(self._app_version)

        # if app_must_update:
        #     self._log.debug("[login] need to update app version")

        user_config: dict = await self._api.user_config()
        self._log.debug(f"[user_config] returned\n{json_pp(user_config)}")
        u_id = user_config["uId"]

        user_verify_email: dict = await self._api.user_verify_email(
            self._username, self._client_id
        )
        self._log.debug(f"[user_verify_email] returned\n{json_pp(user_verify_email)}")
        code: str = user_verify_email["code"]
        id: str = user_verify_email["id"]

        salted: str = salted_hash(self._password, u_id)
        self._log.debug(f"[login] using salted hash {salted}")

        encrypted: str = encrypted_hash(code, salted)
        self._log.debug(f"[login] using encrypted hash {encrypted}")

        user_login: dict = await self._api.user_login(self._username, encrypted, id)
        self._log.debug(f"[user_login] returned\n{json_pp(user_login)}")

        self._device_id = user_login["deviceId"]
        self._token = user_login["token"]
        self._refresh_token = user_login["refreshToken"]
        self._api.set_bearer_token(self._refresh_token)

    async def refresh_token(self) -> None:
        """FIXME: user_refresh_token doesn't seem to work once token is expired..."""
        if self._token is None:
            raise UhooError("Error cannot refresh token prior to logging in")

        try:
            user_refresh_token: dict = await self._api.user_refresh_token(
                self._token, self._device_id
            )

            self._log.debug(
                f"[user_refresh_token] returned\n{json_pp(user_refresh_token)}"
            )

            self._token = user_refresh_token["token"]
            self._refresh_token = user_refresh_token["refreshToken"]
            self._api.set_bearer_token(self._refresh_token)

        except UnauthorizedError:
            self._log.debug(
                "\033[91m"
                + "[refresh_token] received 401 error, attempting to re-login"
                + "\033[0m"
            )
            await self.login()

    async def get_latest_data(self) -> None:
        try:
            data_latest: dict = await self._api.data_latest()
        except UnauthorizedError:
            self._log.debug(
                "\033[93m"
                + "[get_latest_data] received 401 error, refreshing token and trying again"
                + "\033[0m"
            )
            await self.refresh_token()
            data_latest = await self._api.data_latest()
        except ForbiddenError:
            self._log.debug(
                "\033[93m"
                + "[get_latest_data] received 403 error, refreshing token and trying again"
                + "\033[0m"
            )
            await self.refresh_token()
            data_latest = await self._api.data_latest()

        # self._log.debug(f"[data_latest] returned\n{json_pp(data_latest)}")

        self.user_settings_temp = data_latest["userSettings"]["temp"]

        device: dict
        for device in data_latest["devices"]:
            serial_number: str = device["serialNumber"]
            if serial_number not in self._devices:
                self._devices[serial_number] = Device(device)

        for data in data_latest["data"]:
            serial_number = data["serialNumber"]
            device_obj: Device = self._devices[serial_number]
            if device_obj.timestamp < data["timestamp"]:
                device_obj.update_data(data)

    def get_device(self, serial_number) -> Optional[Device]:
        if serial_number in self._devices:
            return self._devices[serial_number]
        else:
            return None

    def get_devices(self) -> Dict[str, Device]:
        return self._devices



USER_AGENT_PRODUCT: str = "uHoo"
USER_AGENT_PRODUCT_VERSION: str = "10.3.1"
USER_AGENT_SYSTEM_INFORMATION: str = "iPhone;13.3; iOS 14.7.1; Scale/3.00"

APP_VERSION: int = 93
CLIENT_ID: str = "0000000000000000000|0000000000000000000"



from typing import Any


class Device(object):
    def __init__(self, device: dict) -> None:
        self.calibration: int
        self.created_at: str  # "YYYY-MM-DDThh:mm:ss.sssZ" UTC time
        self.home: Any  # type unknown
        self.latitude: Any  # type unknown
        self.longitude: Any  # type unknown
        self.mac_address: str
        self.name: str
        self.serial_number: str
        self.server: int
        self.ssid: str
        self.status: int

        self.co: float  # might be an int
        self.co2: float  # might be an int
        self.dust: float
        self.humidity: float
        self.no2: float
        self.ozone: float
        self.pressure: float
        self.temp: float
        self.timestamp: int
        self.voc: float  # might be an int

        self.update_device(device)
        self.timestamp = -1

    def __repr__(self):
        return f"<{self.__module__}.{self.__class__.__name__} serial_number: {self.serial_number!r}>"

    def update_device(self, device: dict) -> None:
        self.calibration = device["calibration"]
        self.created_at = device["createdAt"]
        self.home = device["home"]
        self.latitude = device["latitude"]
        self.longitude = device["longitude"]
        self.mac_address = device["macAddress"]
        self.name = device["name"]
        self.serial_number = device["serialNumber"]
        self.server = device["server"]
        self.ssid = device["ssid"]
        self.status = device["status"]

    def update_data(self, data: dict) -> None:
        self.co = data["co"]["value"]
        self.co2 = data["co2"]["value"]
        self.dust = data["dust"]["value"]
        self.humidity = data["humidity"]["value"]
        self.no2 = data["no2"]["value"]
        self.ozone = data["ozone"]["value"]
        self.pressure = data["pressure"]["value"]
        self.temp = data["temp"]["value"]
        self.timestamp = data["timestamp"]
        self.voc = data["voc"]["value"]



import asyncio
import logging

import click
from aiohttp import ClientSession

from pyuhoo import Client


async def init_client(username, password, websession) -> Client:
    return Client(username, password, websession, debug=False)


async def example(username, password):
    async with ClientSession() as websession:
        client: Client = await init_client(username, password, websession)

        print("[-] Logging in")
        await client.login()
        print("[+] Successfully logged in")

        print("[-] Getting latest data")
        await client.get_latest_data()
        print("[+] Successfully received data")

        devices = client.get_devices()
        device = next(iter(devices.values()))  # get reference to first device

        print("[+] Data from first device:")
        print(
            f"    Serial Number: {device.serial_number}\n"
            + f"    Timestamp:     {device.timestamp}\n"
            + f"    Temperature:   {device.temp} {client.user_settings_temp}\n"
        )

        while True:
            delay = 60 * 5
            print(f"[-] Sleeping for {delay} seconds...")
            await asyncio.sleep(delay)

            print("[-] Getting latest data")
            await client.get_latest_data()  # polls/updates all devices
            print("[+] Successfully received data")

            print("[+] Data from first device:")
            print(
                f"    Serial Number: {device.serial_number}\n"
                + f"    Timestamp:     {device.timestamp}\n"
                + f"    Temperature:   {device.temp} {client.user_settings_temp}\n"
            )


@click.command()
@click.option("--username", "-u", prompt=True, help="uHoo username")
@click.option(
    "--password",
    "-p",
    prompt=True,
    hide_input=True,
    confirmation_prompt=False,
    help="uHoo password",
)
@click.option("--debug", is_flag=True)
def cli(username, password, debug):
    if debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.WARN)

    asyncio.get_event_loop().run_until_complete(example(username, password))


if __name__ == "__main__":
    cli()
    print("[+] Done!")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyuhoo']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.7.4,<4.0.0', 'click>=8.0.1,<9.0.0', 'pycryptodome>=3.10.1,<4.0.0']

entry_points = \
{'console_scripts': ['pyuhoo-cli = pyuhoo.cli:cli']}

setup_kwargs = {
    'name': 'pyuhoo',
    'version': '0.0.5',
    'description': 'Python API for talking to uHoo consumer API',
    'long_description': '# pyuhoo\n\n[![PyPi version](https://img.shields.io/pypi/v/pyuhoo.svg)](https://pypi.python.org/pypi/pyuhoo/)\n[![pre-commit.ci status](https://results.pre-commit.ci/badge/github/csacca/pyuhoo/master.svg)](https://results.pre-commit.ci/latest/github/csacca/pyuhoo/master)\n![ci workflow](https://github.com/csacca/pyuhoo/actions/workflows/ci.yaml/badge.svg)\n\nPython API for talking to uHoo consumer API\n\nPlease note that this is a non-public API that has been reverse-engineered from mobile\napps. It is likely to break unexpectedly when uHoo changes the API.\n',
    'author': 'Christopher Sacca',
    'author_email': 'csacca@csacca.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/csacca/pyuhoo',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
